""" Vindby
    Copyright (C) 2018  Esther Dornhelm

    This program is licensed under an Attribution-NonCommercial 4.0 International (CC BY-NC 4.0) license

    Terms and conditions of this license are defined in
    file 'LICENSE.txt', which is part of this source code package.

    See <https://creativecommons.org/licenses/by-nc/4.0/>."""


import keyboard, timeit, mOptimizer, mInteraction, mGameTools, GUI, mMarket
from mStakeholder import *
from mWindFarmGrid import *
from mWallet import *
from mSeaGrid import *
from mRepairVessel import *
from mWindFarm import *
from mCosts import *
from mCO2 import *
from mEnergyDemand import *
from mOMoptions import *
from mOandM import *
from mTurbineAvailable import *
from mSubstructureAvailable import *
from mReport import *


class Game:
    """Represents the game"""
    global key
    global keypressed
    def __init__(self):                             #Establish attributes of the game
        self.userName=None
        self.seaGrid = SeaGrid()                    #Create one copy of sea grid
        self.tn=None                                #End date
        self.gameClock=None                         #Timer
        self.windClock=None
        self.windFarmGrid=WindFarmGrid()
        self.wallet=Wallet()
        self.availableTurbines=TurbineAvailable()
        self.availableSubstructures=SubstructureAvailable()
        self.failureRates=FailureRates()
        self.costs=Costs()
        self.energyDemand=EnergyDemand(self.seaGrid.windInt)
        self.renewEnergySupply = {'Total':0,'Current':0}
        self.co2 = CO2(self.energyDemand,self.renewEnergySupply)
        self.OMoptions = OMoptions()
        self.interact = mInteraction.Interaction(self)  # Give the game to the user object, which interacts and processes user input
        self.settings={'RegularReports':'Quinquennial','GameSpeed_minutes':1440*7}
        self.tool=mGameTools.gameTools(self)
        self.timerInfo=0
        self.optimizer=mOptimizer.Optimizer(self)
        self.market = mMarket.market()
        self.goals =[]
        self.myGoal = None
        self.simulationTimer = 1
        self.paused = False
        self.quit = False

    def setup(self):                                #Setup the game by importing data, creating objects
        self.seaGrid.read()
        self.tn=525600*1000+1                       #Ending time - to update
        self.gameClock=0                            #Starting time. game timer can be chosen by user to run faster or slower, but wind simulations are still done at every windtime timestep.
        self.windClock=0                            #wind clock is the timer for wind simulations, which is always every 10 mintes
        self.costs.read()
        self.repairVessel = RepairVessel()
        self.availableTurbines.read()              #reads turbines available at start of the game. updates with achievements
        self.availableSubstructures.read()
        self.failureRates.read()
        self.failureRates.sort()
        self.govt=Government()
        self.public=Public()
        self.shareholders=Shareholders()
        self.market.setup(self.seaGrid,self.availableTurbines)
        self.interact.intro()
        self.tool.setup()
        self.gui=GUI.App(self)
        self.menu="\nMAIN MENU\nAt any time enter: \n\t[W]ind - do research, build farms, make changes to existing wind farms\n\t[R]eport - Generate a report showing details of the game's status\n\t[L]edger - view a detailed list of your expenses and profits\n\t[F]ailures - view and repair current failures in your windfarms\n\t[S]ettings - change game speed and/or frequency of auto generated reports\n\t[P]ause\n\t[Q]uit\n"
        print(self.menu)

    def kcallback(self,x):                          #detect key press from user during simulation
        global key
        global keypressed
        if keypressed:
            return
        key= x.name
        print('You entered: ',key)
        print('\n...Wait for simulation to finish...\n')
        keypressed=True

    def playgame(self):                             #Main Game Loop running while waiting for keyboard input in the background
        global key
        global keypressed
        global khook
        # global quit                                 #Not ideal, but leave for now
        self.quit=False
        keypressed=False
        key=""
        khook=keyboard.on_press(self.kcallback,True) #can supress certain combinations such as ctrl+c, or to quit game, try keyboard.block_key
        import time

        startReportgetNumbers = timeit.default_timer()

        while self.gameClock<=self.tn and self.quit==False: #This controls how fast the game is running for user experience
            time.sleep(self.simulationTimer)                               #run game at slow or fast speeds. number of seconds between game pause. simulations still run at set tick level (10 min or 1 hour)
            startSimulationTimer = timeit.default_timer() #Must keep!
            while self.windClock <=self.gameClock:      #This loop is controlled by the wind clock, which represents the duration of one tick in game calcs. for example, the wind speed is calculated every hour regardless of how fast the game clock is running
                if self.windClock % self.tool.reportIntervals[self.settings['RegularReports']] == 0 and self.windClock != 0:  # Check whether it is time to publish the annual, monthly, or weekly report
                    r = Report(self, self.settings['RegularReports'])  # Issue a report for this time of game
                    r.issue()
                renewableEnergyAllFarms=0
                farm_weather_time={}
                for farm in self.windFarmGrid.windFarmList:
                    farm.status['age']=self.windClock-farm.status['birthday']
                    if farm.status['age']==farm.lifetime and farm.status['phase']!='Decommissioned':#Wind Farm has reached normal or extended lifetime
                        stopReportgetNumbers = timeit.default_timer()
                        print('20 year simulation of your wind farm: ',stopReportgetNumbers-startReportgetNumbers)
                        self.pause()
                        self.interact.endofLifetime(farm)
                        self.unpause()
                    elif farm.status['age']>0 and farm.status['phase']=='Operating': #farm is operating, running normally!
                        run=farm.run(self.windClock,self.seaGrid.windInt) #Run wind farm and collect / distribute energy based on what month it is, and the wind clock resolution (1 hour)
                        farm_weather_time[farm.name]=run['simulation time']
                        farm.OandM.maintain() #Check for failures.
                        output=run['energy']
                        renewableEnergyAllFarms+=output
                        co2j=self.co2.CO2prevented(output)
                        farm.co2prevented_byfarm+=co2j
                        self.tool.breakEven(farm, 'co2BreakEven') #Check whether farm has broke even on co2 emitted
                        if self.windClock % 43800 == 0 and self.windClock != 0:  # Once a month, recalculate LCOE per farm
                            farm.updateLCOE()
                        revenue = farm.sellingPrice['First10Years'] * output
                        farm.revenue += round(revenue,2)
                        self.tool.breakEven(farm,'CostBreakEven')  # Check whether farm has broke even on CAPEX cost yet
                        self.tool.breakEven(farm,'ProfitBreakEven')  # Check whether farm has broke even on CAPEX cost yet
                        self.wallet.deposit(revenue,str(self.tool.convertTime(self.windClock)['Short Date']) + ' : Energy revenue from ' + farm.name + ' @ ' + str(run['windSpeed']) + ' m/s and ' + str(run['waveHeight']) + ' m')
                    elif farm.status['age']<0: #Wind Farm is being constructed
                        farm.build(self.costs.costlist['contractor_perday'],self.seaGrid.windInt)
                        if farm.farmFailure!={} and farm.notification['fail']=='new':      #If there is an accident during construction, add to total list of failures
                            self.windFarmGrid.listFailures(farm,farm.myCell,farm.farmFailure[farm])
                            self.shareholders.meeting(str(farm.name + 'Construction accident causing delays'), -2)
                            farm.notification['fail']='old'
                            self.gui.updateCellColor(farm.myCell)
                    elif farm.status['age']>=0 and farm.status['phase']=='Under Construction': #Wind Farm has finsihed construction
                        print('\n{} Construction complete! Time to commission.'.format(farm.name))
                        farm.status['phase']='Operating'
                        self.windFarmGrid.installedCapacity += farm.installedCapacity
                        farm.costs['CAPEX'] += farm.costs['contractorCost']
                        print("\tTotal contractor fees: €{:5,.2f} ".format(farm.costs['contractorCost']))
                        self.wallet.withdraw(farm.costs['contractorCost'],str(self.tool.convertTime(self.windClock)['Short Date'])+' : Contractor paid for construction of Farm '+str(farm.name))
                        print('\tExceeded schedule by {:1.1f} days'.format((farm.constructiontime['actual']-farm.constructiontime['original'])/1440))
                        co2i = self.co2.CO2fromConstruction(len(farm.myTurbineList))  # emit co2 from manufactoring and transportation
                        farm.co2install = co2i  # Go into the negative in terms of c02 emission. break even by saving co2 from non-renewables
                        farm.updateLCOE()
                        print('\tCO2 emitted during manufacturing and transport of wind farm [million T]: {}\n'.format(round(co2i, 2)))
                        self.optimizer.optimizeStructures(farm,farm.myTurbineList[0].type,farm.mySubstructureList[0].type)
                        self.tool.checkDidYouDoYourResearch(farm) #Check for things that could've been prevented from research
                        self.gui.updateCellColor(farm.myCell)

                self.renewEnergySupply['Total'] += renewableEnergyAllFarms
                self.renewEnergySupply['Current'] = renewableEnergyAllFarms
                self.energyDemand.renewableShare = self.windFarmGrid.installedCapacity * 24 * 365 / self.energyDemand.demandYearly #Need to update as farms are being commissioned/decommissioned
                self.co2.energytoCO2()              #calculate how much co2 has been emitted due to production from renewables, and remainder from renewables
                self.windClock+=self.seaGrid.windInt
                self.tool.checkRewards()
                self.tool.updateScore()
                if self.windClock % (525600/2) ==0 and self.windClock != 0: #Every six months, issue a short "optimizing report" for O&M
                    self.tool.checkCF()
                if self.windClock % 525600 == 0 and self.windClock != 0:  #Every year, adjust prices for inflation
                    self.optimizer.report()
                    self.costs.inflation()
                    self.energyDemand.update()  # Update market demand for energy

                self.checkkeypress() #Check if user has pressed any keys
                if self.wallet.checkbalance() <= 0:
                    print('\n/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/Game Over!\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/')
                    print(":o !! Looks like you've run out of funds and you're bankrupt... until I add the feature to take out loans and pay them back.... its GAME OVER. Try again!")
                    self.quit = True
                if self.quit==True: #If user has decided to end the game
                    r=Report(self,'FINAL')
                    r.issue()
                    self.tool.endgame()
                    break
            self.gui.update()
            self.gameClock+=self.settings['GameSpeed_minutes']
            result = self.tool.checkGoal()
            if result == 'Loss':
                self.endGame()

            ####### To test simulation times, use this section###########
            # stopSimulationTimer = timeit.default_timer()
            # simulationTime = stopSimulationTimer - startSimulationTimer
            # print('one game loop : ', simulationTime)
            #
            # for farm,t in farm_weather_time.items():
            #     print('Farm {} weather simulation in sec: {}'.format(farm,t))
            # self.simulationTimer = max(0,1-simulationTime)
            # print('sleep time ',self.simulationTimer)
            #
            # actual_speed = self.settings['GameSpeed_minutes']/(self.simulationTimer + simulationTime)
            # print('actual speed in game minutes per second: ',actual_speed)
            #########################################################

        if self.gameClock >= self.tn:
            r = Report(self,'FINAL')
            r.issue()
            self.endGame()

    def pause(self):
        global khook
        keyboard.unhook(khook)
        self.paused = True
        self.gui.update()

    def unpause(self):
        global khook #keypressed = False
        khook = keyboard.on_press(self.kcallback, True)
        self.paused = False
        self.gui.update()

    def roundTo(self,value,nearest):
        return int(math.ceil(value / nearest)) * nearest

    def checkkeypress(self):
        global key
        global keypressed
        global khook
        if keypressed and self.quit == False:
            keyboard.unhook(khook)
            self.paused = True
            self.gui.update()
            if key.lower() =="w":
                start_build = timeit.default_timer()
                windFarmInput=self.interact.designwindfarm()    #Get input for designing a wind farm
                try:
                    w=WindFarm(self.availableTurbines.turbineList, self.availableSubstructures.substructureList)
                    w.design(**windFarmInput)
                    w.OandM=OMmanager(windFarmInput['OMstrategy'],w,self)
                    w.birthday=self.windClock
                    w.costs['CAPEX'] = windFarmInput['numturbines'] * (windFarmInput['turbineunitcost'] + windFarmInput['structureunitcost']) + (self.costs.costlist['cable_export_unitm'] * w.cableLength * 1000) + (self.costs.costlist['cable_interarray_unitm'] * w.interarray_cableLength) + sum(windFarmInput['Additional_Costs'].values())
                    substation_cost = 0
                    if self.windFarmGrid.onshoreSubstation.exist == False: #For first onshore substation
                        print('\n\t>>Before installing any offshore cabling, you first will need to build an onshore substation. To accommodate future connections, you must build a 500 MW facility at a cost of: €{:,}'.format(self.roundTo(self.costs.costlist['substation_onshore'],1000)))
                        self.windFarmGrid.onshoreSubstation.build()
                        substation_cost += self.costs.costlist['substation_onshore']
                    while self.windFarmGrid.onshoreSubstation.capacity < self.windFarmGrid.installedCapacity + w.installedCapacity:#Each time another 500MW capacity is added, add another onshore substation
                        print('\n\t>>The addition of this farm will exceed the capacity of the onshore substation. Now you must build another 500 MW facility at a cost of: €{:,}'.format(self.roundTo(self.costs.costlist['substation_onshore'],1000)))
                        self.windFarmGrid.onshoreSubstation.addCapacity()
                        substation_cost += self.costs.costlist['substation_onshore']
                    if w.substation == 'offshore':
                        print('\n\t>>Distance to onshore connection is {}. An offshore substation is required at a cost of €{:,} to setup an HVDC connection, which minimizes power losses.'.format(w.cableLength, self.roundTo(self.costs.costlist['substation_offshore'],1000)))
                        substation_cost += self.costs.costlist['substation_offshore']

                    if w.costs['CAPEX'] + substation_cost <self.wallet.balance and w.costs['CAPEX']>0 and windFarmInput['numturbines']!=0:
                        news = False #Use to print the notification about AC/DC cablse just once
                        try:
                            min=w.cableLength
                            for otherfarm in self.windFarmGrid.availableConnections:        #Check if there are farms closer to the substation that we can connect to
                                cnxn=math.sqrt(math.pow((otherfarm.myCell.xcoor-w.myCell.xcoor),2)+math.pow((otherfarm.myCell.ycoor-w.myCell.ycoor),2))
                                if cnxn<min:
                                    if w.substation == 'onshore':
                                        min=round(math.sqrt(math.pow((otherfarm.myCell.xcoor-w.myCell.xcoor),2)+math.pow((otherfarm.myCell.ycoor-w.myCell.ycoor),2)),2)
                                        farmcnxn=otherfarm
                                    elif w.substation == 'offshore' and otherfarm.substation == 'offshore':
                                        min = round(math.sqrt(math.pow((otherfarm.myCell.xcoor - w.myCell.xcoor), 2) + math.pow((otherfarm.myCell.ycoor - w.myCell.ycoor), 2)), 2)
                                        farmcnxn = otherfarm
                                    elif w.substation == 'offshore' and otherfarm.substation == 'onshore':
                                        if news == False:
                                            print('There are existing farms closer to your farm, but you can not connect to their system because you need DC subsea cables, and they use AC.')
                                            news = True #Only print this once per design session
                            if min!=w.cableLength:
                                potential_savings= int(self.costs.costlist['cable_export_unitm'] * (w.cableLength - min) * 1000)
                                connect=input('\nCable length to the onshore substation is {} km, but you can connect to {} using {} km of cable. You save €{:,} in cost, but if {} has a grid failure, so does your new farm. Would you like to choose this option? Y/N'.format(w.cableLength,farmcnxn.name,min,potential_savings,farmcnxn.name))
                                if connect=='Y' or connect=='y':
                                    w.cableLength=min
                                    w.connection = farmcnxn                                 #make sure the farm knows about which other farm its connected to.
                                    self.windFarmGrid.connections[w]=w.connection           #Tell wind farm grid about the new connections
                                    w.costs['CAPEX'] -= potential_savings                   #incorporate savings into total cost
                        except:
                            pass
                        #Caclulate the exact failure rate for this farm given the cable length
                        w.myFailureRates.failureRates['farm_grid'] = w.cableLength * w.myFailureRates.failureRates['farm_grid_perkm'] #Get failure rate for this farm's subsea cables
                        w.myFailureRates.farmFailureRates['farm_grid'] = w.myFailureRates.failureRates['farm_grid']
                        del w.myFailureRates.farmFailureRates['farm_grid_perkm'] #Delete the placeholder for the unit failure rate. Replaced by actually as calculated for this farm's cable length
                        w.sellingPrice['First10Years'] = self.market.sellingPrice(w)
                        print('\n\t>>Based on your design, the grid operator is offering to pay a fixed price of €{} per MWhr for your energy. Your profit will be this selling price - the Levelized Cost of Energy (LCOE). Maximize profit by reducing downtime and capital costs'.format(w.sellingPrice['First10Years']))
                        w.OandM.setup(w)
                        construct=input('\nDo you want to build this wind farm for €{:,} Y/N? Construction time estimated at {} days including fabrication and installation. '.format(self.roundTo(w.costs['CAPEX'],100000),round(w.constructiontime['original']/1440,0)))

                        if construct=='y' or construct=='Y':
                            farm_name = self.interact.namewindfarm(w)
                            w.status['phase']='Under Construction'
                            w.status['birthday']=self.windClock+w.constructiontime['original'] #Assumed date of farm completion base don proposed schedule. can be delayed
                            w.status['age']=-w.constructiontime['original'] #Assume scheduled construction time, assigned as a negative value. delays can occur
                            w.costs['contractorCost']=0 #hire contractor
                            w.name = farm_name.lower()
                            self.windFarmGrid.add(w)  # Add the constructed windfarm to the grid - now it will run during the simulation
                            self.wallet.withdraw(w.costs['CAPEX'],str(self.tool.convertTime(self.windClock)['Short Date']) + ' : Investment in building Wind ' + w.name)
                            self.seaGrid.occupy(w.myCell,w)  # Label cell as occupied. no more farms in this cell
                            self.gui.updateCellColor(w.myCell)
                            self.gui.updateCellName(w.myCell)
                    elif windFarmInput['numturbines']==0:
                        print('No turbines in wind farm.')
                    else:
                        print('\n\t>>Capital Cost of wind farm: \t\t€{:,}'.format(self.roundTo(w.costs['CAPEX'],1000)))
                        self.wallet.display()
                        print('\n**********************************:/***************************************')
                        print("********** Insufficient funds to build both the wind farm and the substation... save some more money!! **********")
                        print('**********************************:/***************************************\n')
                except:                             #If the user exits interact.designwindfarm and does not want to build, or other error
                    pass
                stop_build = timeit.default_timer()
                time_to_build=stop_build-start_build
                self.timerInfo+=time_to_build
            if key == "L" or key == "l":
                self.wallet.displayLedger()
                input('\n::Game is paused:: Enter any key to return to MAIN MENU: ')
            if key == "S" or key == "s":
                self.interact.adjustsettings()
            if key == "R" or key == "r":
                r = Report(self, 'On Demand')  # Issue a report for this time of game
                r.issue()
                ans=input('\n::Game is paused:: Enter any key to continue (go into [S]ettings in the main menu to adjust frequency of regular reports): ')
            if key=="F" or key=="f":
                if self.windFarmGrid.failureList!=[]:
                    repair_session_time=self.gameClock#Start a repair session for each wind farm
                    while self.windFarmGrid.failureList!=[]:
                        self.windFarmGrid.displayFailures(self.costs.costlist)
                        # reaction = self.interact.managefailuresinput()
                        toRepairorNot, repairIndex = self.interact.managefailuresinput()
                        if toRepairorNot == True and repairIndex <len(self.windFarmGrid.failureList):
                            failure_to_repair = self.windFarmGrid.failureList[repairIndex] #Find the failure object to be repaired
                            failure_to_repair[0].OandM.manageFailures(failure_to_repair, repairIndex ,repair_session_time) #Go into the farm's O&M manager and manage the repair
                        else:
                            break
                else:
                    input('\nNo failures to display. ::Game is paused:: Enter any key to return to MAIN MENU: ')
            if key == "p" or key == "P":
                input('::Game is paused:: Enter any key to return to MAIN MENU')
            if key.lower() != 'q':
                print(self.menu)
            if key=="q" or key=="Q":
                yayornay = input('Are you sure you want to quit the game? Y/N')
                if yayornay.lower() == 'y':
                    self.quit=True
            keypressed = False
            khook = keyboard.on_press(self.kcallback, True)
            self.paused = False
            self.gui.update()

        if self.quit == True:
            self.endGame()

    def endGame(self):
        global key
        global keypressed
        global khook
        self.quit = True
        self.tool.endgame()
        keyboard.unhook(khook)
        import sys
        sys.exit(0)


# ------------------------------Run the program---------------------------------


x=Game()
startSetup = timeit.default_timer()
x.setup()
stopSetup=timeit.default_timer()
startGame = timeit.default_timer()
x.playgame()
stopGame = timeit.default_timer()


print('Setup time: [sec]',round(stopSetup-startSetup,2))
print('Total Game real Time: [sec]', round(stopGame - startGame,2))
print('Total Game Building real Time [sec]:', round(x.timerInfo,2))
print('Game Time simulated: [year] ', round(x.gameClock/525600,2))
import sys
sys.exit(0)