""" Vindby
    Copyright (C) 2018  Esther Dornhelm

    This program is licensed under an Attribution-NonCommercial 4.0 International (CC BY-NC 4.0) license

    Terms and conditions of this license if defined in
    file 'LICENSE.txt', which is part of this source code package.

    See <https://creativecommons.org/licenses/by-nc/4.0/>."""

import numpy

class CO2:
    """Represents CO2 in the worlds atmosphere"""

    def __init__(self,energyDemand,renewableEnergy):            #Give the energy demand of the game to the CO2 scientist
        self.energy=energyDemand
        self.renewableEnergySupply=renewableEnergy
        self.conc=[]                                            #Total CO2 in the atmosphere at a certain time
        self.emissions = {'Current':0,'Total':0}                #Emissions accumulate per  time intervals, without intervention
        self.emissions_prevented = 0                            #Cumulative CO2 in MT prevented by using wind farms
        self.emissions_construction=[]                          #CO2 emitted per construction of each farm
        self.ppm=400
        self.nonEnergyppm =  0.0000014
        self.temp=0                                             #Temperature increase
        self.nonrenewableRate=0.00000095                        #Mt CO2eq per MWhr, offshore wind the facts chapter 6
        self.renewableRate =  0.000000012                       #Mt CO2eq per MWhr
        self.homes=0
        self.maxhomes = 0

    def energytoCO2(self):
        nonrenewableEnergy = max(0,self.energy.demandHourly - self.renewableEnergySupply['Current']) #The different between the enrgy demand and that suspplied by the wind farms is assumed to be covered by non renewables
        CO2_Nonrenew=nonrenewableEnergy*self.nonrenewableRate
        CO2_Renew=self.renewableEnergySupply['Current']*self.renewableRate
        CO2_saved=self.renewableEnergySupply['Current']*(self.nonrenewableRate-self.renewableRate)
        CO2_released=CO2_Nonrenew+CO2_Renew
        absorbed=CO2_released/2                                 #Assume about half of what is emitted is absorbed naturally
        emitted=CO2_released-absorbed
        self.emissions['Current'] = emitted
        self.emissions['Total'] += emitted
        self.emissions_prevented += CO2_saved
        self.updatetemp()
        self.homescalc()

    def CO2prevented(self,farmoutput):
        CO2prev=farmoutput*(self.nonrenewableRate-self.renewableRate)
        return CO2prev

    def updatetemp(self): #Temperature increase
        self.nonEnergyppm *= (1 + self.energy.increaseRate) ** (1 / (24 * 365))     #Additional ppm from other things happening in the world
        self.ppm += self.nonEnergyppm + (self.emissions['Current'] / 1000) / 7.81   #7.81 Gigton Co2 is 1 ppm
        self.temp=self.ppm*(2/(415-400))+(2-415*(2/(415-400)))

    def homescalc(self):
        self.homes = round(self.renewableEnergySupply['Current'] / (3.9 / 8760),0)  #homes on avg 3.9 MWhr per home per year. this value is updated every hour
        if self.homes > self.maxhomes:
            self.maxhomes = self.homes

    def CO2fromConstruction(self,n):                                                #Calculate the emissions from manufacturing and isntalling the OWF based on n=number of turbines
        co2_construction=.020000 +5*n*(.000769+.000123)                             #values From lit review. estimates tCO2 per ton of steel and for ship transport between USA/China - see backup in report
        self.emissions_construction.append(co2_construction)
        self.emissions['Total'] += co2_construction
        return co2_construction

    def display(self):
        print('\tHomes that can be powered by offshore wind:\t {:,}'.format(self.maxhomes))
        print('\tPercent of demand met by renewables:\t\t{:2.3f}%'.format(self.energy.renewableShare*100))
        print('\tCO2 prevented to date* [MT]:\t\t\t\t{:3,.2f}'.format(self.emissions_prevented))
        print('\t\t*Equivalent to {:2,.0f} km driven by a passenger car'.format(3000*self.emissions_prevented))
        print('\t\t*Equivalent to carbon sequestered by {:2,.1f} acres of forest in one year'.format(0.8 * self.emissions_prevented))
        print('\tCurrent CO2 in atmosphere[ppm]:\t\t\t\t{:3,.1f}'.format(self.ppm))
        print('\tIncrease in world temperature [deg]:\t\t+{:2.2f}'.format(self.temp))


