""" Vindby
    Copyright (C) 2018  Esther Dornhelm

    This program is licensed under an Attribution-NonCommercial 4.0 International (CC BY-NC 4.0) license

    Terms and conditions of this license are defined in
    file 'LICENSE.txt', which is part of this source code package.

    See <https://creativecommons.org/licenses/by-nc/4.0/>."""

import time, pandas, csv
from mReport import *

class gameTools():
    """Represents tools used by game so as not to crowd the game file"""

    def __init__(self,game):
        self.g=game
        self.reportIntervals={}
        self.news = {}
        self.header={}
        self.d=time.strftime('%Y%m%d')
        self.t=time.strftime("%H%M%S", time.localtime())
        self.rewardMeasures={'homes':1000000,'turbine_farmrequirement':5,'structure_farmrequirement':7,'CF_min':0.3,'CF_max':0.7}      #Dictionary that stores the evolving thresholds needs to reach to earn a reward. E.g. number of homes powered must increase by some amount each time the previous goal is reached.
        self.rewardGranted={'homesPowered':False,'unlocksubstructure':False}
        self.scores = {'Profitability':0,'Climate Impact': 0,'Stakeholders':0, 'Capacity Factor':0 ,'Total': 0}
        self.RLfarms = []
        self.goal_items = []
        self.result = 'Still Playing'

    def setup(self):
        self.regularReports()               #Load frequency of annual, monthly, and weekly reports
        with open('tfarms.txt', 'r') as f:  #Load existing OWF names
            for line in f:
                line = line.rstrip('\n')
                line = line.lower()
                self.RLfarms.append(line)
        self.goal_items = [(self.g.wallet.profits - self.g.wallet.operatingCosts) / self.g.wallet.initialInvestment,self.g.tool.scores['Total'], self.g.energyDemand.renewableShare * 100, self.g.co2.temp, 0]

    def convertTime(self, clock):           #convert game clock to calendar date. days per month isnt ctually correct here
        yr = math.floor(clock/(60*24*365))
        mon = math.floor(((clock/(60*24*365))-yr)*12)
        day= math.floor(((((clock/(60*24*365))-yr)*12)-mon)*(365/12))
        hr=math.floor(((((((clock/(60*24*365))-yr)*12)-mon)*(365/12))-day)*24)
        min=math.floor(((((((((clock/(60*24*365))-yr)*12)-mon)*(365/12))-day)*24)-hr)*60)
        calendar_year = 2018 + yr
        calendar_month = mon + 1
        date='Date: ' + str("%04d" % calendar_year)+'-'+str("%02d" % calendar_month)+'-'+str("%02d" % day)+' Time: '+str("%02d" % hr)+':'+str("%02d" % min)
        shortdate = str("%04d" % calendar_year)+'-'+str("%02d" % calendar_month)+'-'+str("%02d" % day)
        return {'Date':date, 'Short Date':shortdate,'minutes':min,'hours':hr,'days':day,'months':mon,'years':yr,'calendar year': calendar_year}

    def regularReports(self):               #input time intervals for regularly issued reports, or other uses
        self.reportIntervals['Quinquennial'] = 525600*5
        self.reportIntervals['Annual']=525600
        self.reportIntervals['Monthly'] = 525600 / 12

    def newspaper(self,news,time):          #use one line function to publish news the first time it happens, if it keeps happening
        toprint=time+news
        if toprint not in self.news:
            print(news)
            self.news[toprint] = 'oldnews'

    def checkDidYouDoYourResearch(self,farm):
        messages = []
        if farm.myCell.env_protection == 'Sensitive' and farm.status['Environmentally Mitigated'] == False:
            farm.myFailureRates.update('farm_stakeholder_public', 10)  # increases failure rate for public shutdown
            self.g.public.meeting(str(farm.name + 'Building in an environmentally sensitive area'), -3)
            self.g.govt.meeting(str(farm.name + 'Building in an environmentally sensitive area'), -3)
            printMe ="!!You've built a wind farm on an environmentally SENSITIVE area and you did not do any mitigation during construction! The public is very upset and youll have to pay a fine of €{:,}. You should have done some research! ".format(self.roundTo(self.g.costs.costlist['fine_environmental_violation'], 1000))
            messages.append(printMe)
            self.g.wallet.withdraw(self.roundTo(self.g.costs.costlist['fine_environmental_violation'],1000), self.convertTime(self.g.windClock)['Short Date'] + ' : Constructing in an environmentally sensitive area without noise mitigating measures in Cell ' + str(farm.myCell.id))
        if farm.myCell.env_protection == 'Protected':
            self.g.public.meeting(str(farm.name + 'Building in an environmentally protected area'), -5)
            self.g.govt.meeting(str(farm.name + 'Building in an environmentally protected area'), -5)
            self.g.shareholders.meeting(str(farm.name + 'Building in an environmentally protected area, and loosing everyones money'), -10)
            printMe ="!!You've built a wind farm in a PROTECTED environmental area. This is really bad and illegal. The public is not happy. You're going to have to decommission the farm and pay a gnarly €{:,} fine. And no, you will not get your money back. You should have done some research! ".format(self.roundTo(self.g.costs.costlist['fine_environmental_violation'], 1000))
            messages.append(printMe)
            self.g.interact.decommissionWindFarm(farm)
            self.g.wallet.withdraw(self.g.costs.costlist['fine_environmental_violation'], self.convertTime(self.g.windClock)['Short Date'] + ' : Illegally building in an environmentally protected area in Cell ' + str(farm.myCell.id))
        elif farm.myCell.shipping_channel == 'Active':
            self.g.govt.meeting(str(farm.name + 'Building through a shipping route'), -2)
            printMe = "!!You've build your wind farm through a major shipping channel, now vessels have to reroute around your farm. The public and the government is not happy, and now you have to pay a €{:,} fine. You should have done some research! ".format(self.roundTo(self.g.costs.costlist['fine_reroute_vessels'], 1000))
            messages.append(printMe)
        if farm.myCell.dist2shore <= 13:  # If you 've built really close to shore
            printMe = "!!The residents living at the shoreline don't want to see massive structures in their backyard. A distance under 13km is considered a major visual effect. The public is not happy and want this farm shut down."
            messages.append(printMe)
            farm.myFailureRates.update('farm_stakeholder_public', 10)  # increases failure rate for public shutdown
            self.g.public.meeting(str(farm.name + ' Building way too close to shore'), -5)
        elif farm.myCell.dist2shore <= 24:  # If you 've built really close to shore
            printMe = "!!The residents living at the shoreline don't want to see massive structures in their backyard, even at a distance. A distance between 13 and 24 km is considered a moderate visual effect. The public is displeased."
            messages.append(printMe)
            farm.myFailureRates.update('farm_stakeholder_public', 10)  # increases failure rate for public shutdown
            self.g.public.meeting(str(farm.name + ' Building close to shore'), -3)

        for message in messages:
            print('\n**********************************>:(***************************************')
            print('*****' + str(message) + '*****')
            print('************************************>:(*************************************\n')

    def checkRewards(self):
        messages = []
        #Check if its time to unlock more turbines!
        if sum(1 for farm in self.g.windFarmGrid.windFarmList if farm.status['phase'] == 'Operating') >= self.rewardMeasures['turbine_farmrequirement'] and self.g.availableTurbines.unlock!=0: #Unlocks turbines if there are at least five (starting threshold) operating wind farms and if there are still turbines left to unlock
        #STUBBED OUT - CREATE AN R&D OBJECT TO WHICH THE LAYER CAN INVEST MONEY INTO, OR AUTOMIATICALLY PUT MONEY INTO ONCE IN A WHILE. THEEN USE THAT AS A CHECK HERE TO RELEASE MORE TURBINES
            printMe = "All this offshore wind farm building has been pushing industry to catch up. Now there are even larger turbines available to choose from."
            messages.append(printMe)
            self.g.availableTurbines.read()
            self.rewardMeasures['turbine_farmrequirement'] += 5 #After another 5 farms are built, unclock another turbine if there are any more to unlock

        #Check if its time to release the floating substructure
        if sum(1 for farm in self.g.windFarmGrid.windFarmList if farm.status['phase'] == 'Operating') >= self.rewardMeasures['structure_farmrequirement'] and self.g.availableSubstructures.unlock!=0:
            # STUBBED OUT - CREATE AN R&D OBJECT TO WHICH THE LAYER CAN INVEST MONEY INTO, OR AUTOMIATICALLY PUT MONEY INTO ONCE IN A WHILE. THEEN USE THAT AS A CHECK HERE TO RELEASE MORE STRUCTURES
            printMe = "All this offshore wind farm building has been pushing industry to catch up. Now there are new substructures available to choose from."
            messages.append(printMe)
            self.g.availableSubstructures.read()
            self.rewardMeasures['structure_farmrequirement'] += 3 # After another 5 farms are built, unclock another substructure if there are any more to unlock
            self.rewardGranted['unlocksubstructure'] = True

        #Check if we reached another 100,000 homes powered
        if self.g.co2.homes > self.rewardMeasures['homes']:
            self.g.govt.meeting(str('Increasing capacity to another 1,000,000 households'),1)
            self.g.public.meeting(str('Increasing capacity to another 1,000,000 households'), 1)
            print('Another 1,000,000 homes are being powered by offshore wind energy!\n')
            self.rewardMeasures['homes']+=1000000
            self.rewardGranted['homesPowered']=True

        for message in messages:
            print('\n**********************************:D***************************************')
            print('*****' + str(message) + '*****')
            print('************************************:D*************************************\n')

    def checkCF(self): #Check if capacity factor of any farm is below a threshold
        for farm in self.g.windFarmGrid.windFarmList:
            if farm.status['phase']=='Operating':
                if farm.capacityFactor <= self.rewardMeasures['CF_min']:
                    self.g.shareholders.meeting(str('{} Capacity factor is below {}'.format(farm.name,self.rewardMeasures['CF_min'])),-.5)
                elif farm.capacityFactor >=  self.rewardMeasures['CF_max']:
                    self.g.shareholders.meeting(str('{} Capacity factor is above {}'.format(farm.name,self.rewardMeasures['CF_max'])), +.5)

    def checkGoal(self):
        if int(self.g.tool.convertTime(self.g.windClock)['calendar year']) >= self.g.myGoal['year'] and self.result == 'Still Playing':
            goal_items = [(self.g.wallet.profits - self.g.wallet.operatingCosts) / self.g.wallet.initialInvestment,self.g.tool.scores['Total'], self.g.energyDemand.renewableShare * 100, self.g.co2.temp, 0]
            n = self.g.myGoal['index']
            print('\n/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/Time is up!\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/')
            if self.g.myGoal['measure'] == 'Renewable Energy Share [%]':
                if goal_items[n] >= self.g.myGoal['target']:
                    print("\n\n\n\nYou've reached your goal, nice! Keep playing to achieve another 30% (higher level feature to be added in future versions of game, for now close and play again. )")
                    self.result = 'Win'
                    try:
                        self.g.pause()
                        a = input('Enter any key to keep playing: ')
                        self.g.unpause()
                    except:
                        print('Did you know? The global growth rate of offshore wind energy is about 30% each year in installed capacity.')
                else:
                    print("\n\n\n\nYou didn't reach your goal in time :( Build more capacity to get higher RES share. Build efficient farms to get more profit so you can build more.")
                    self.result = 'Loss'
            if self.g.myGoal['measure'] == 'Global Temperature Increase [degC] +':
                if goal_items[n] < self.g.myGoal['target']:
                    print("\n\n\n\nYou've reached your goal, nice! Keep playing to keep the temperature from reaching X by XXXX (higher level feature to be added in future versions of game, for now close and play again.)")
                    self.result = 'Win'
                    try:
                        self.g.pause()
                        a = input('Enter any key to keep playing: ')
                        self.g.unpause()
                    except:
                        print('Did you know? The first commercial offshore wind farm was named Vindeby in the Baltic Sea.')
                else:
                    print("\n\n\n\nYou didn't reach your goal in time :( The only way to decrease temperatures is to lower CO2 emissions. Produce more wind energy with more capacity and high capacity factors")
                    self.result = 'Loss'
            if self.g.myGoal['measure'] == 'Profit Over Initial Investment Ratio':
                if goal_items[n] >= self.g.myGoal['target']:
                    if self.g.wallet.capitalCosts >= self.g.wallet.initialInvestment:
                        print("\n\n\n\nYou've reached your goal, nice! Here's another zillion dollars to invest (higher level feature to be added in future versions of game, for now close and play again.)")
                        self.result = 'Win'
                        try:
                            self.g.pause()
                            a = input('Enter any key to keep playing: ')
                            self.g.unpause()
                        except:
                            print('Did you know? Wind turbines can be taller than the Statue of Liberty.')
                    else:
                        print("\n\n\n\nYou've broke profit on all investments, but you didn't spend all the money you were given. Next time, spend all you got!")
                        self.result = 'Loss'
                else:
                    print("\n\n\n\nYou didn't reach your goal in time :( Next time, make more investments, spend all the money you have, but wisely.")
                    self.result = 'Loss'
            if self.g.myGoal['measure'] == 'Overall Score [/100]':
                print("\n\n\n\nTimes up! So, how'd you do? The person with the lowest score has to buy everyone a beer.")
                self.result = 'Loss'
            if self.g.myGoal['measure'] == 'Free Play':
                print("\n\n\n\nTimes up! Hope you enjoyed the game! Hopefully in the future this free play can be multi-player, and on a platform where you can save and return to your game.")
                self.result = 'Loss'
        return self.result

    def breakEven(self, farm, request): #EVENTUALLY, MOVE THIS TO MGAMETOOLS.PY
        if  (farm.costs['CAPEX'] ) < farm.revenue and request == 'CostBreakEven' and farm.breakEven['cost'] == False:
            payback_period=self.g.windClock-farm.status['birthday'] #in hours
            print("\n :D !!! Congrats! You've paid back wind farm {} in {} days!".format(farm.name, round(payback_period/1440, 0)))
            self.g.wallet.deposit(10000,str(self.g.tool.convertTime(self.g.windClock)['Short Date'])+' : Bonus for paying back wind farm '+str(farm.name)+' for '+str(farm.costs['CAPEX']))
            farm.breakEven['cost']=True
            self.g.shareholders.meeting(str(farm.name + ' Break even on Farm capital cost'),5)
        if farm.co2prevented_byfarm > farm.co2install and request == 'co2BreakEven' and farm.breakEven['co2'] == False:
            payback_period = self.g.windClock - farm.status['birthday']  # in hours
            print("\n :D !!! Congrats! You've broke even on CO2 emissions for your wind farm {} in {} days!".format(farm.name, round(payback_period/1440, 0)))
            farm.breakEven['co2']=True
            self.g.govt.meeting(str(farm.name + ' Break even on CO2 emissions'), 3)
            self.g.public.meeting(str(farm.name + ' Break even on CO2 emissions'),1)
            self.g.wallet.deposit(10000,str(self.g.tool.convertTime(self.g.windClock)['Short Date'])+' : Bonus for breaking even on CO2 emissions at wind farm '+str(farm.name)+' for '+str(farm.co2install)+' MT')
        if (farm.LCOE.value) < farm.sellingPrice['First10Years'] and request == 'ProfitBreakEven' and farm.breakEven['Profit'] == False:
            print("\n :D !!! Congrats! You're bringing in a profit at farm {}! This means the LCOE is lower than the price of energy".format(farm.name))
            self.g.wallet.deposit(10000, str(self.g.tool.convertTime(self.g.windClock)['Short Date']) + ' : Bonus for breaking profit at wind farm ' + str(farm.name) + 'with LCOE of €' + str(farm.LCOE.value) +'/MWh')
            farm.breakEven['Profit'] = True
            self.g.shareholders.meeting(str(farm.name + ' Started generating profit'), 3)

    def updateScore(self):
        P = 25 #points per category
        try:
            self.scores['Profitability'] = round(min(P,((self.g.wallet.profits / self.g.wallet.investments)*(P/1))),1)
        except:
            self.scores['Profitability'] = 0
        self.scores['Climate Impact'] = round((self.g.energyDemand.renewableShare/.3)*P,1) #how close are we to the 30% goal?
        self.scores['Stakeholders'] = P/2 + (max(0,min(P/6,self.g.govt.satisfaction)) +
                                        max(0,min(P/6,self.g.public.satisfaction)) +
                                        max(0,min(P/6,self.g.shareholders.satisfaction)))
        sum = 0
        N = 0
        for farm in self.g.windFarmGrid.windFarmList:
            sum += farm.capacityFactor
            N += 1
        for farm in self.g.windFarmGrid.retiredWindFarms:
            sum += farm.capacityFactor
            N += 1
        try:
            self.scores['Capacity Factor'] = P * (sum / N)
        except:
            self.scores['Capacity Factor'] = 0
        self.scores['Total'] = self.scores['Profitability'] + self.scores['Climate Impact'] + self.scores['Stakeholders'] + self.scores['Capacity Factor']

    def roundTo(self,value,nearest):
        return int(math.ceil(value / nearest)) * nearest

    def endgame(self): #Use this to close all the record files for the overall game and individual wind farms
        filename = './GamesPlayed/' + str(self.g.interact.user) + 'Game Results_' + str(self.d) + '_' + str(self.t) + '_Overall.csv'  # File with records about          entire game
        filename_farm = './GamesPlayed/' + str(self.g.interact.user) + 'Game Results_' + str(self.d) + '_' + str(self.t) + '_Farms.csv'  # File with records about entire game
        final_results = {}
        final_results['start_time'] = self.t
        final_results['end_time'] = time.strftime("%H%M%S", time.localtime())
        final_results['game_duration'] = self.g.gameClock
        final_results['Game_Played'] = self.g.myGoal['measure']
        goal_item = self.goal_items[self.g.myGoal['index']]
        final_results['Goal_progress'] = goal_item
        final_results['Result'] = self.result
        final_results['Balance'] = self.roundTo(self.g.wallet.checkbalance(), 100000)
        final_results['Capital_costs'] = self.roundTo(self.g.wallet.capitalCosts, 100000)
        final_results['Operating_costs'] = self.roundTo(self.g.wallet.operatingCosts, 100000)
        final_results['Revenue'] = self.roundTo(self.g.wallet.profits, 100000)
        final_results['Installed_cap'] = self.g.windFarmGrid.installedCapacity
        final_results['Grid_Connections'] = len(self.g.windFarmGrid.connections)
        inv_cells =[]
        for cell in self.g.interact.investigatedCells:
            inv_cells.append(cell.id)
        final_results['Investigated_cells'] = inv_cells
        final_results['public_score'] = self.g.public.satisfaction
        final_results['govt_score'] = self.g.govt.satisfaction
        final_results['shareholder_score'] = self.g.shareholders.satisfaction
        final_results['RES_share'] = self.g.energyDemand.renewableShare*100
        final_results['CO2 prevent'] = self.g.co2.emissions_prevented
        final_results['Number_farms'] = len(self.g.windFarmGrid.windFarmList) + len(self.g.windFarmGrid.retiredWindFarms)

        f = open(filename, 'w')
        for key in final_results.keys():
            item = key
            value = final_results[key]
            row = str(item) + "," + str(value) + "\n"
            f.write(row)
        f.close()

        farm_results = {}
        farm_headers = ['Farm_name','Info','N_Failures_farm','N_Failures_structures','N_Failures_turbines','DT_farm','DT_structures','DT_turbines']

        f_farm = open(filename_farm,'w')
        writer = csv.DictWriter(f_farm, fieldnames=farm_headers, delimiter=',', lineterminator='\n')
        writer.writeheader()

        for farm in self.g.windFarmGrid.windFarmList:
            farm_entry = {}
            farm_entry = {'name': str(farm.name), 'Cell':str(farm.myCell.id),'CAPEX':str(farm.roundTo(farm.costs['CAPEX'], 100000)),
                          'Turbine':str(farm.myTurbineList[0].type.cap),'Foundation':str(farm.mySubstructureList[-1].type.type),
                          'No_turbines':str(len(farm.myTurbineList)),'Cable_Length':str(farm.cableLength),'O&M_strategy':str(farm.OandM.strategy),
                          'Capacity_Factor':str(farm.capacityFactor),'Revenue':str(farm.roundTo(farm.revenue, 100000)),
                          'Status':str(farm.status['phase']),'Birthday':str(farm.status['birthday']),'Age':str(farm.status['age']),'Lifetime':str(farm.lifetime)}
            farm_results['Farm_name'] ='windfarm_'+str(farm.name)
            farm_results['Info'] = farm_entry
            farm.OandM.endGame()
            farm_results['N_Failures_farm']= farm.OandM.reportNumberFailures['farm']
            farm_results['N_Failures_structures']=farm.OandM.reportNumberFailures['structure']
            farm_results['N_Failures_turbines']=farm.OandM.reportNumberFailures['turbine']
            farm_results['DT_farm']=farm.OandM.reportDownTime['farm']
            farm_results['DT_structures']=farm.OandM.reportDownTime['structure']
            farm_results['DT_turbines']=farm.OandM.reportDownTime['turbine']

            writer.writerow(farm_results)

        for farm in self.g.windFarmGrid.retiredWindFarms:
            farm_entry = {}
            farm_entry = {'name': str(farm.name), 'Cell':str(farm.myCell.id),'CAPEX':str(farm.roundTo(farm.costs['CAPEX'], 100000)),
                          'Turbine':str(farm.myTurbineList[0].type.cap),'Foundation':str(farm.mySubstructureList[-1].type.type),
                          'No_turbines':str(len(farm.myTurbineList)),'Cable_Length':str(farm.cableLength),'O&M_strategy':str(farm.OandM.strategy),
                          'Capacity_Factor':str(farm.capacityFactor),'Revenue':str(farm.roundTo(farm.revenue, 100000)),
                          'Status':str(farm.status['phase']),'Birthday':str(farm.status['birthday']),'Age':str(farm.status['age']),'Lifetime':str(farm.lifetime)}
            farm_results['Farm_name'] = 'windfarm_' + str(farm.name)
            farm_results['Info'] = farm_entry

            farm.OandM.endGame()
            farm_results['N_Failures_farm'] = farm.OandM.reportNumberFailures['farm']
            farm_results['N_Failures_structures'] = farm.OandM.reportNumberFailures['structure']
            farm_results['N_Failures_turbines'] = farm.OandM.reportNumberFailures['turbine']
            farm_results['DT_farm'] = farm.OandM.reportDownTime['farm']
            farm_results['DT_structures'] = farm.OandM.reportDownTime['structure']
            farm_results['DT_turbines'] = farm.OandM.reportDownTime['turbine']

            writer.writerow(farm_results)

        print('\nWe hope you enjoyed the game. Goodbye!\n')
        print('Please close the windows (cmd window and interface) to exit and start a new game.')
