""" Vindby
    Copyright (C) 2018  Esther Dornhelm

    This program is licensed under an Attribution-NonCommercial 4.0 International (CC BY-NC 4.0) license

    Terms and conditions of this license are defined in
    file 'LICENSE.txt', which is part of this source code package.

    See <https://creativecommons.org/licenses/by-nc/4.0/>."""

import math

class Interaction:
    """Compilation of user interaction methods"""

    def __init__(self,game): #interface must have a reference to the game
        self.g=game
        self.user=None
        self.investigatedCells=[]

    def intro(self):
        self.g.goals = ["Invest all of your initial investment (€1 bn) in capital costs, and break overall profit by 2025.",
                        "Keep playing until 2030 and achieve an overall score higher than the person next to you.",
                        "Achieve 10% share of all energy supply can be provided by offshore wind by 2050.",
                        "Prevent the global temperature from increasing by 1 degree by 2050.",
                        "Free play until 2100 or whenever you want to stop."]
        goal_measures = ['Profit Over Initial Investment Ratio','Overall Score [/100]','Renewable Energy Share [%]','Global Temperature Increase [degC] +','Free Play']
        goal_year = [2025,2030,2050,2050,2100]
        goal_target = [1,100,10,1,0]
        goal_initial_investment = [1e9,1e9,10e9,10e9,10e9]

        print('_________________WELCOME TO THE GAME VINDBY!_________________\n')
        print("Vindby is a world with a growing energy demand, rising temperatures, and great big windy sea. You've been given an initial budget to invest in offshore wind energy.\n"
              "Here are the basics of what you need to know:\n"
              "\t>> Offshore wind farms convert wind power offshore in electric power onshore, improving the climate by replacing nonrenewable sources.\n"
              "\t>> Offshore wind can be very costly compared to other energy sources. Knowing where and how to cut costs is crucial.\n"
              "\t>> Given the amount of investment, international carbon reduction targets, and growing public concern, stakeholder satisfaction can make or break a project.\n"
              "\t>> Wind energy is not constant because the wind is not constant. Maximize output by minimizing downtimes due to failures.\n")
        self.user =input("Enter your name (no special characters e.g. !@#$ ): ")
        print("Choose your goal that you would like to achieve:\n")
        for g in range(len(self.g.goals)):
            print('\t({}) {}'.format(g+1,self.g.goals[g]))
        myGoal_i = input('My goal: ')

        while myGoal_i != '1' and myGoal_i != '2' and myGoal_i != '3' and myGoal_i != '4' and myGoal_i != '5':
            myGoal_i = input('Enter a goal 1 through 5: ')
        n = int(myGoal_i)-1
        self.g.wallet.deposit(goal_initial_investment[n], str(self.g.tool.convertTime(self.g.windClock)['Short Date']) + ' : Initial Deposit')
        self.g.myGoal = {'index':n,'Description':self.g.goals[n],'year':goal_year[n],'measure':goal_measures[n],'target':goal_target[n]} #'item':goal_items[n],
        print('\nGreat, lets play!')

    def designwindfarm(self):
        print('\n_________________DESIGN A WIND FARM_________________\n')
        R=input('The Sea Grid is divided into 100 cells, and there is an onshore substation located near cell 1. Choose a cell 1-100 to build on, type [R] to do research for {:,} per cell, type [S] to see all investigations to date, [U] to update the O&M stragety for an existing wind farm,or [E]xit:\n'.format(self.roundTo(self.g.costs.costlist['Field_investigation'],1000)))
        while True: #isinstance(int(R),int)==False:
            try:
                if isinstance(int(R),int)==True:
                    break
            except:
                if R=='R' or R=='r':
                    if self.g.wallet.balance>=self.g.costs.costlist['Field_investigation']:
                        c=101 #slightly unusual loop to account for letter input
                        while int(c) not in range(1, (len(self.g.seaGrid.cellList)) + 1):
                            try:
                                while int(c) not in range(1, (len(self.g.seaGrid.cellList)) + 1):
                                    c = input('\tChoose do do research on cells between 1 and {}: '.format((len(self.g.seaGrid.cellList))))
                                break
                            except:
                                c = input('\tPlease enter a cell between 1 and {}: '.format((len(self.g.seaGrid.cellList))))
                                c=101
                        cell=self.g.seaGrid.cellList[int(c)-1]
                        print('\n',cell)
                        self.investigatedCells.append(cell)
                        self.g.gui.updateCellColor(cell)
                        self.g.wallet.withdraw(self.g.costs.costlist['Field_investigation'],str(self.g.tool.convertTime(self.g.windClock)['Short Date'])+' : Field Investigation of '+str(cell.id))
                        self.g.gui.update()
                    else:
                        print('Not enough moneys')
                elif R=='S' or R=='s':
                    print('\nAll investigation reports:\n')
                    for cell in self.investigatedCells:
                        print(cell)
                elif R=='e' or R=='E':
                    break
                elif R.lower() == 'u':
                    WF = input('Which farm would you like to update? Enter name:')
                    if str(WF) in self.g.windFarmGrid.windfarmnames:
                        for farm in self.g.windFarmGrid.windFarmList:
                            if farm.name == str(WF):
                                print('You want to edit ', farm.name)
                                ans = self.YesorNo('Do you want to modify the O&M strategy? Y/N')
                                if ans.lower() == 'y':
                                    self.g.OMoptions.displayOptions(self.g.costs.costlist,self.g.failureRates.otherfailures['conditionbasedmonitoring_efficiency'])
                                    OMindex = int(input('Enter choice of Maintenance Strategy:'))  # Get maintenance strategy
                                    while OMindex > len(self.g.OMoptions.strategyList) or OMindex == 0:
                                        OMindex = int(input('Enter choice of strategy from the list:'))
                                    old_strategy = str(farm.OandM.strategy)
                                    new_strategy = self.g.OMoptions.strategyList[OMindex - 1]
                                    farm.OandM.strategy = new_strategy
                                    print('{} strategy changed from {} to {}'.format(farm.name,old_strategy,farm.OandM.strategy))
                    else:
                        print('Farm {} not found.'.format(str(WF)))
                R=input('\nChoose a cell (1-100) to build on, do [R]esearch, view the [S]eagrid for investigations, [U]pdate existing farm O&M, or [E]xit.')
        try:
            myCellID=int(R)
            while myCellID not in range(0,(len(self.g.seaGrid.cellList))+1) or self.g.seaGrid.cellList[myCellID-1].status=='occupied':
                if myCellID not in range(0,(len(self.g.seaGrid.cellList))+1):
                    myCellID = int(input('There is no Cell {} in the Sea Grid. Choose another.'.format(myCellID)))
                elif self.g.seaGrid.cellList[myCellID-1].status=='occupied':
                    myCellID = int(input('There is already a wind farm in Cell {}. Choose another.'.format(myCellID)))

            x=self.g.availableTurbines
            x.display(self.g.costs.costlist)
            turbineIndex=int(input('Enter choice of turbine:'))
            while turbineIndex > len(x.turbineList) or turbineIndex ==0:
                turbineIndex = int(input('Enter choice of turbine from the list:'))

            y=self.g.availableSubstructures
            y.display(self.g.costs.costlist, self.g.seaGrid.cellList[myCellID-1].depth, self.g.seaGrid.cellList[myCellID-1].soilqual)
            substructureIndex=int(input('Enter choice of sub-structure: (Cost per structure for location given soil type and depth. Includes substructure, foundaion, and transition piece.)'))
            while substructureIndex > len(y.substructureList) or substructureIndex ==0:
                substructureIndex = int(input('Enter choice of sub-structure from the list:'))

            cont = False
            while cont == False:
                numTurbines = input('Enter number of turbines: ')
                if isinstance(int(numTurbines),int) == True:
                    if int(numTurbines) > 100 or int(numTurbines) < 0:
                        print('Cell is large enough for 100 turbines maximum.')
                    else:
                        numTurbines = int(numTurbines)
                        cont = True
            self.g.OMoptions.displayOptions(self.g.costs.costlist,self.g.failureRates.otherfailures['conditionbasedmonitoring_efficiency'])
            OMindex = int(input('Enter choice of Maintenance Strategy:'))  # Get maintenance strategy
            while OMindex > len(self.g.OMoptions.strategyList) or OMindex == 0:
                OMindex = int(input('Enter choice of strategy from the list:'))
            strategy = self.g.OMoptions.strategyList[OMindex - 1]

            turbinecap= x.turbineList[turbineIndex-1].cap   #Get capacity of this turbine to find and save the unit cost (done here because windfarm input limited to **windfarminput)
            lookup = str(turbinecap) + 'mwturbine'
            turbineunitcost =self.g.costs.costlist[lookup]

            structureunitcost = y.substructureList[substructureIndex-1].priceQuote(self.g.costs.costlist, self.g.seaGrid.cellList[myCellID-1].depth, self.g.seaGrid.cellList[myCellID-1].soilqual)

            additional_costs={}
            scour_protection = self.YesorNo('Do you want to install scour protection Y/N? Scour is the loss of sediment '
                                            'around the base of bottom founded substructures and occurs in '
                                            'areas of poor / loose soil.''\n\tCost per substructure now: €{:,}'
                                            '\n\tCost per substructure after construction if needed: €{:,}'.format(self.roundTo(self.g.costs.costlist['scour_protection_design_perturbine'],100000),self.roundTo(self.g.costs.costlist['structure_localscour'],100000)))
            if scour_protection.lower() == 'y':
                additional_costs['Scour_protections'] = self.g.costs.costlist['scour_protection_design_perturbine'] * numTurbines
            else:
                additional_costs['Scour_protections'] = 0

            mitigation = self.YesorNo('Do you want to implement noise mitigation measures during construction Y/N? Noise mitigation is required in environmentally sensitive areas where marine mammals live and are sensitive to loud noises. (Cost: {:,})'.format(self.roundTo(self.g.costs.costlist['environmental_mitigation'],100000)))
            if mitigation.lower() == 'y':
                additional_costs['Environmental Mitigation'] = self.g.costs.costlist['environmental_mitigation']
            else:
                additional_costs['Environmental Mitigation'] = 0

            return {'cell':self.g.seaGrid.cellList[myCellID-1],'turbineIndex':turbineIndex,'numturbines':numTurbines,'turbineunitcost':turbineunitcost,'structureunitcost':structureunitcost,'substructureIndex':substructureIndex,'OMstrategy':strategy,'Additional_Costs':additional_costs}
        except:
            print('build next time!')

    def namewindfarm(self,farm):
        farm_name = input('Great, name your new farm: ').lower()
        if farm_name.lower() in self.g.windFarmGrid.windfarmnames:
            while farm_name in self.g.windFarmGrid.windfarmnames:
                farm_name = input('That name is already taken, choose another: ')
        if farm_name.lower() in self.g.tool.RLfarms:
            print("Oh hey! You seem to know a bit about the real life market. Since you're an insider, lets see if we can speed up the construction process.")
            farm.constructiontime['original'] = 1440 * 2
            farm.constructiontime['actual'] = 1440 * 2
        return farm_name

    def managefailuresinput(self):
        rxn = input('\nEnter the list number of the failure to repair or any other key to continue with game.')# Costs will be {} for the trip to the farm, plus {} per repair.\n'.format(self.g.costs.costlist['repairVesselCharter'],self.g.costs.costlist['full_repair']))
        try:
            rxn = int(rxn) - 1 #Check that input is integer
            toRepair=True       #Tell game to issue repair
        except:
            toRepair = False    #Don't do any repairs
        return [toRepair,rxn] #tell game to repair!

    def endofLifetime(self,farm):
        print('\nWind {} has reached its {:2.0f} year lifetime.'.format(farm.name,farm.lifetime/525600))
        if farm.status['life']=='normal' or farm.status['life']=='extended5yr' :
            ans=input('Choose to [D]ecommission or [E]xtent lifetime by 5 years. '
                      '\nDecommissioning requires a large operation (Cost: €{:,} per structure and €{:,} per km of cable '
                      'if not connected to anther farm).\nRecycling the farm can get you back €{:,} per structure.'
                      '\nHowever, extending the lifetime now doubles the failure rate. Lifetime can be extended '
                      'a maximum of two times (ten years).'.format(self.roundTo(self.g.costs.costlist['dismantle_farm_perMW'] * farm.myTurbineList[0].type.cap,1000),self.roundTo(self.g.costs.costlist['dismantle_cable_perkm'],1000),
                                                                   self.roundTo(self.g.costs.costlist['dismantle_recycle_farm_perMW'] * farm.myTurbineList[0].type.cap,1000)))
            while ans!='D' and ans!='d' and ans!='E' and ans!='e':
                ans=input('Please choose to [D]ecommission or [E]xtent lifetime by 5 years.')
            ans = 'd' #Temporary assign decomission to all farms - just for results collection for report
            if ans=='e' or ans=='E':
                farm.lifetime+=5*525600
                print('Lifetime extended by 5 years.')
                if farm.status['life']=='normal':
                    farm.status['life']='extended5yr'
                elif farm.status['life']=='extended5yr':
                    farm.status['life']='extended10yr'
                for turbine in farm.myTurbineList:
                    for failure, rate in turbine.myTurbineFailureRates.items():
                        rate *= 1.5 #failure rate increases by 50% when lifetime is extended

            if ans=='d' or ans=='D':
                self.decommissionWindFarm(farm)
        elif farm.status['life']=='extended10yr':
            print("You've reached the maximum lifetime of the farm of {:2.0f} years! No more extensions. Time to decommission.".format(farm.lifetime))
            self.decommissionWindFarm(farm)

    def decommissionWindFarm(self,farm):
        print('Decommissioning time! ')
        farm.status['phase'] = 'Decommissioned'
        self.g.gui.updateCellColor(farm.myCell)
        self.g.seaGrid.unoccupy(farm.myCell) #remove from seagrid, but keep in windfarmgrid for record
        self.g.windFarmGrid.retireFarm(farm) #dismantle wind farm, but move the information to another list. game does not run these wind farms.
        farm.costs['Dismantle'] = self.g.costs.costlist['dismantle_farm_perMW'] * farm.installedCapacity
        farm.status['Dismantle'] = 'Dismantle farm' #keep track of what is being dismantled (farm and cables or just farm)
        if farm not in self.g.windFarmGrid.connections.values(): #If there are no connections to this farm relying on transmission, remove cable. Otherwise, leave cable.
            self.g.windFarmGrid.availableConnections.remove(farm)
            farm.costs['Dismantle']+= self.g.costs.costlist['dismantle_cable_perkm'] * farm.cableLength
            farm.status['Dismantle'] = 'Dismantle farm and subsea cables'
        farm.costs['Recycle']= self.g.costs.costlist['dismantle_recycle_farm_perMW'] * farm.installedCapacity #Get some money back for recycling
        self.g.wallet.withdraw(farm.costs['Dismantle'], self.g.tool.convertTime(self.g.windClock)['Short Date'] + ' : ' + str(farm.status['Dismantle']))
        self.g.wallet.deposit(farm.costs['Recycle'], self.g.tool.convertTime(self.g.windClock)['Short Date'] + ' : Recycling payback')
        print(farm.costs)
        print(farm.status)

    def adjustsettings(self):
        print('Current game settings:')
        d={}
        i=1
        for setting, stat in self.g.settings.items():
            d[i]=setting #Used to index settings and change when prompted
            print('\t({}) {} : {}'.format(i, setting, stat))
            i += 1
        print('\t({}) Keep all defaults'.format(i))
        ans=input('Choose a setting you would like to edit.')
        if ans == 'selfdestructin54321': #Used to induce errors during debugging
            print("\nI've placed a weakness deep within the system... a flaw so small and powerful, they will never find it. \n")
            i='induce error'
            j = i + 1
        elif ans == 'rosebud':
            self.g.wallet.deposit(1000000000,'fresh from the laundromat')
        try:
            if int(ans)-1 <= len(self.g.settings)+1:
                if d[int(ans)] == 'GameSpeed_minutes': #if user wants to adjust game speed
                    t = 'None'
                    while t.lower() != 's' and t.lower() != 'n' and t.lower() != 'f':# and t.lower() != 'm':
                        t = input('Enter desired game speed:\n\t[S]low : 1 hour per second\n\t[N]ormal : 1 week per second (Default)\n\t[F]ast : 1 month per second')# \n\t[M]ax simulation speed')
                    if t.lower() == 's':
                        self.g.settings['GameSpeed_minutes'] = 60
                    if t.lower() == 'n':
                        self.g.settings['GameSpeed_minutes'] = 1440*7
                    if t.lower() == 'f':
                        self.g.settings['GameSpeed_minutes'] = 43800
                elif d[int(ans)] == 'RegularReports':
                    a=input('Choose [Q]uinquennial (every five years), [A]nnual, or [M]onthly (default is Quinquennial)')
                    if a.lower() == "a":
                        self.g.settings['RegularReports'] = 'Annual'
                    if a.lower() == "q":
                        self.g.settings['RegularReports'] = 'Quinquennial'
                    if a.lower() == "m":
                        self.g.settings['RegularReports']= 'Monthly'
        except:
            pass

    def YesorNo(self,Question):
        answer=input(Question)
        while answer.lower() != 'y' and answer.lower() != 'n':
            answer=input('Please answer [Y]es or [N]o: ')
        return answer

    def roundTo(self,value,nearest):
        return int(math.ceil(value / nearest)) * nearest

