""" Vindby
    Copyright (C) 2018  Esther Dornhelm

    This program is licensed under an Attribution-NonCommercial 4.0 International (CC BY-NC 4.0) license

    Terms and conditions of this license are defined in
    file 'LICENSE.txt', which is part of this source code package.

    See <https://creativecommons.org/licenses/by-nc/4.0/>."""

import math
from mTurbineAll import *

class TurbineAvailable:
    """List of available turbines"""

    def __init__(self):
        self.allTurbines = None #object with all turbines. Used to pass info to market
        self.turbineList=[]
        self.powerCurveList=[]
        self.count=3            #Number of available turbines that the game begins with
        self.unlock=0           #Assign below - the number of turbines that can be unlocked.

    def read(self):
        self.allTurbines = TurbineAll()
        self.allTurbines.read()
        self.turbineList = []
        self.unlock = len(self.allTurbines.turbineList) - self.count   #Number of remaining turbines to be unlocked. At zero, game stops checking
        for i in range(0,self.count):
            self.turbineList.append(self.allTurbines.turbineList[i])
            self.powerCurveList.append(self.allTurbines.powerCurveList[i])
        self.count+=1           #Each time the function read is called, class sets aside the next item in the list to be "unlocked" in the game.

    def grab(self,i):
        return self.turbineList[i]

    def display(self,costs):
        i=1
        print('Turbines available are:')
        for turbine in self.turbineList:
            lookup = str(turbine.cap) + 'mwturbine'
            cost = self.roundTo(int(costs[lookup]),100000)
            print('\t({}) {} MW (Cost: €{:,} each)'.format(i,turbine.cap,cost))
            i+=1

    def roundTo(self,value,nearest):
        return int(math.ceil(value / nearest)) * nearest

#x=TurbineAvailable()
#x.read(3)
#x.display()
